/* === This file is part of Calamares - <http://github.com/calamares> ===
 *
 *   Copyright 2014, Aurélien Gâteau <agateau@kde.org>
 *
 *   Calamares is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   Calamares is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Calamares. If not, see <http://www.gnu.org/licenses/>.
 */

#include <PartitionCoreModule.h>

#include <CreatePartitionJob.h>
#include <CreatePartitionTableJob.h>
#include <DeletePartitionJob.h>
#include <FormatPartitionJob.h>
#include <DeviceModel.h>
#include <PartitionInfo.h>
#include <PartitionModel.h>
#include <PMUtils.h>
#include <Typedefs.h>
#include <utils/Logger.h>

// CalaPM
#include <CalaPM.h>
#include <core/device.h>
#include <core/partition.h>
#include <backend/corebackend.h>
#include <backend/corebackendmanager.h>

// Qt
#include <QStandardItemModel>

class PartitionIterator
{
public:
    Partition* operator*() const
    {
        return m_current;
    }

    void operator++()
    {
        if ( !m_current )
            return;
        if ( m_current->hasChildren() )
        {
            // Go to the first child
            m_current = static_cast< Partition* >( m_current->children().first() );
            return;
        }
        PartitionNode* parent = m_current->parent();
        Partition* successor = parent->successor( *m_current );
        if ( successor )
        {
            // Go to the next sibling
            m_current = successor;
            return;
        }
        if ( parent->isRoot() )
        {
            // We reached the end
            m_current = nullptr;
            return;
        }
        // Try to go to the next sibling of our parent

        PartitionNode* grandParent = parent->parent();
        Q_ASSERT( grandParent );
        // If parent is not root, then it's not a PartitionTable but a
        // Partition, we can static_cast it.
        m_current = grandParent->successor( *static_cast< Partition* >( parent ) );
    }

    bool operator==( const PartitionIterator& other ) const
    {
        return m_device == other.m_device && m_current == other.m_current;
    }

    bool operator!=( const PartitionIterator& other ) const
    {
        return ! ( *this == other );
    }

    static PartitionIterator begin( Device* device )
    {
        auto it = PartitionIterator( device );
        PartitionTable* table = device->partitionTable();
        if ( !table )
            return it;
        QList< Partition* > children = table->children();
        // Does not usually happen, but it did happen on a 10MB disk with an MBR
        // partition table.
        if ( children.isEmpty() )
            return it;
        it.m_current = children.first();
        return it;
    }

    static PartitionIterator end( Device* device )
    {
        return PartitionIterator( device );
    }

private:
    PartitionIterator( Device* device )
        : m_device( device )
    {}

    Device* m_device;
    Partition* m_current = nullptr;
};

//- DeviceInfo ---------------------------------------------
PartitionCoreModule::DeviceInfo::DeviceInfo( Device* _device )
    : device( _device )
    , partitionModel( new PartitionModel )
{}

PartitionCoreModule::DeviceInfo::~DeviceInfo()
{
}

void
PartitionCoreModule::DeviceInfo::forgetChanges()
{
    jobs.clear();
    for ( auto it = PartitionIterator::begin( device.data() ); it != PartitionIterator::end( device.data() ); ++it )
        PartitionInfo::reset( *it );
}

//- PartitionCoreModule ------------------------------------
PartitionCoreModule::PartitionCoreModule( QObject* parent )
    : QObject( parent )
    , m_deviceModel( new DeviceModel( this ) )
    , m_bootLoaderModel( new QStandardItemModel( this ) )
{
    // FIXME: Should be done at startup
    if ( !CalaPM::init() )
        qFatal( "Failed to init CalaPM" );

    CoreBackend* backend = CoreBackendManager::self()->backend();
    auto devices = backend->scanDevices();
    for ( auto device : devices )
    {
        auto deviceInfo = new DeviceInfo( device );
        m_deviceInfos << deviceInfo;

        deviceInfo->partitionModel->init( device );
    }
    m_deviceModel->init( devices );

}

PartitionCoreModule::~PartitionCoreModule()
{
    qDeleteAll( m_deviceInfos );
}

DeviceModel*
PartitionCoreModule::deviceModel() const
{
    return m_deviceModel;
}

QAbstractItemModel*
PartitionCoreModule::bootLoaderModel() const
{
    return m_bootLoaderModel;
}

PartitionModel*
PartitionCoreModule::partitionModelForDevice( Device* device ) const
{
    DeviceInfo* info = infoForDevice( device );
    Q_ASSERT( info );
    return info->partitionModel.data();
}

void
PartitionCoreModule::createPartitionTable( Device* device, PartitionTable::TableType type )
{
    DeviceInfo* info = infoForDevice( device );
    // Creating a partition table wipes all the disk, so there is no need to
    // keep previous changes
    info->forgetChanges();

    CreatePartitionTableJob* job = new CreatePartitionTableJob( device, type );
    job->updatePreview();
    info->jobs << Calamares::job_ptr( job );

    refresh( device );
}

void
PartitionCoreModule::createPartition( Device* device, Partition* partition )
{
    auto deviceInfo = infoForDevice( device );
    Q_ASSERT( deviceInfo );

    CreatePartitionJob* job = new CreatePartitionJob( device, partition );
    job->updatePreview();

    deviceInfo->jobs << Calamares::job_ptr( job );

    refresh( device );
}

void
PartitionCoreModule::deletePartition( Device* device, Partition* partition )
{
    auto deviceInfo = infoForDevice( device );
    Q_ASSERT( deviceInfo );

    if ( partition->roles().has( PartitionRole::Extended ) )
    {
        // Delete all logical partitions first
        // I am not sure if we can iterate on Partition::children() while
        // deleting them, so let's play it safe and keep our own list.
        QList< Partition* > lst;
        for ( auto childPartition : partition->children() )
            if ( !PMUtils::isPartitionFreeSpace( childPartition ) )
                lst << childPartition;

        for ( auto partition : lst )
            deletePartition( device, partition );
    }

    QList< Calamares::job_ptr >& jobs = deviceInfo->jobs;
    if ( partition->state() == Partition::StateNew )
    {
        // Find matching CreatePartitionJob
        auto it = std::find_if( jobs.begin(), jobs.end(), [ partition ]( Calamares::job_ptr job )
        {
            CreatePartitionJob* createJob = qobject_cast< CreatePartitionJob* >( job.data() );
            return createJob && createJob->partition() == partition;
        } );
        if ( it == jobs.end() )
        {
            cDebug() << "Failed to find a CreatePartitionJob matching the partition to remove";
            return;
        }
        // Remove it
        if ( ! partition->parent()->remove( partition ) )
        {
            cDebug() << "Failed to remove partition from preview";
            return;
        }
        device->partitionTable()->updateUnallocated( *device );
        jobs.erase( it );
        // The partition is no longer referenced by either a job or the device
        // partition list, so we have to delete it
        delete partition;
    }
    else
    {
        // Remove any PartitionJob on this partition
        for ( auto it = jobs.begin(); it != jobs.end(); )
        {
            PartitionJob* job = qobject_cast< PartitionJob* >( it->data() );
            if ( job && job->partition() == partition )
                it = jobs.erase( it );
            else
                ++it;
        }
        DeletePartitionJob* job = new DeletePartitionJob( device, partition );
        job->updatePreview();
        jobs << Calamares::job_ptr( job );
    }

    refresh( device );
}

void
PartitionCoreModule::formatPartition( Device* device, Partition* partition )
{
    auto deviceInfo = infoForDevice( device );
    Q_ASSERT( deviceInfo );

    FormatPartitionJob* job = new FormatPartitionJob( device, partition );
    deviceInfo->jobs << Calamares::job_ptr( job );

    refresh( device );
}

QList< Calamares::job_ptr >
PartitionCoreModule::jobs() const
{
    dumpQueue();

    QList< Calamares::job_ptr > lst;
    for ( auto info : m_deviceInfos )
        lst << info->jobs;
    return lst;
}

void
PartitionCoreModule::dumpQueue() const
{
    cDebug() << "# Queue:";
    for ( auto info : m_deviceInfos )
    {
        cDebug() << "## Device:" << info->device->name();
        for ( auto job : info->jobs )
            cDebug() << "-" << job->prettyName();
    }
}

void
PartitionCoreModule::refresh( Device* device )
{
    auto model = partitionModelForDevice( device );
    Q_ASSERT( model );
    model->reload();
    updateHasRootMountPoint();
    updateBootLoaderModel();
}

void PartitionCoreModule::updateHasRootMountPoint()
{
    bool oldValue = m_hasRootMountPoint;
    m_hasRootMountPoint = findPartitionByMountPoint( "/" );

    if ( oldValue != m_hasRootMountPoint )
        hasRootMountPointChanged( m_hasRootMountPoint );
}

PartitionCoreModule::DeviceInfo*
PartitionCoreModule::infoForDevice( Device* device ) const
{
    for ( auto deviceInfo : m_deviceInfos )
    {
        if ( deviceInfo->device.data() == device )
            return deviceInfo;
    }
    return nullptr;
}

static QStandardItem*
createBootLoaderItem( const QString& description, const QString& path )
{
    QString text = PartitionCoreModule::tr( "%1 (%2)" )
                   .arg( description )
                   .arg( path );
    QStandardItem* item = new QStandardItem( text );
    item->setData( path, PartitionCoreModule::BootLoaderPathRole );
    return item;
}

void
PartitionCoreModule::updateBootLoaderModel()
{
    m_bootLoaderModel->clear();
    // Can contain up to 2 entries:
    // - MBR of disk which contains /boot or /
    // - /boot or / partition
    QString partitionText;
    Partition* partition = findPartitionByMountPoint( "/boot" );
    if ( partition )
        partitionText = tr( "Boot Partition" );
    else
    {
        partition = findPartitionByMountPoint( "/" );
        if ( partition )
            partitionText = tr( "System Partition" );
        else
            return;
    }
    m_bootLoaderModel->appendRow(
        createBootLoaderItem( tr( "Master Boot Record" ), partition->devicePath() )
    );
    m_bootLoaderModel->appendRow(
        createBootLoaderItem( partitionText, partition->partitionPath() )
    );
}

Partition*
PartitionCoreModule::findPartitionByMountPoint( const QString& mountPoint ) const
{
    for ( auto deviceInfo : m_deviceInfos )
    {
        Device* device = deviceInfo->device.data();
        for ( auto it = PartitionIterator::begin( device ); it != PartitionIterator::end( device ); ++it )
            if ( PartitionInfo::mountPoint( *it ) == mountPoint )
                return *it;
    }
    return nullptr;
}
